img.logo {
  width:100%
}

.right-next {
    float: right;
    max-width: 45%;
    overflow: auto;
    text-overflow: ellipsis;
    white-space: nowrap;
}

.right-next::after{
    content: ' »';
}

.left-prev {
    float: left;
    max-width: 45%;
    overflow: auto;
    text-overflow: ellipsis;
    white-space: nowrap;
}

.left-prev::before{
    content: '« ';
}

.prev-next-bottom {
  margin-top: 3em;
}

.prev-next-top {
  margin-bottom: 1em;
}
